#include <assert.h>
#include <sys/time.h>
#include <stdint.h>

#include <cuda_runtime.h>
#include <cublas.h>
#include <cusparse.h>

#include <bits/mex_interface.h>
#include <bits/mex_convert.h>
#include <bits/timing.h>

#include "matlab_gpu.h"
#include "../src/debug.h"

double *matlab_vector_to_gpu(const mxArray *src)
{
  int n;
  size_t svec;
  double *fsrc, *dst;

  n = mxGetNumberOfElements(src);
  svec = sizeof(double) * n;

  // Copy to double array
  fsrc = (double *)malloc(svec);
  mxArrayToDoubleArray(src, fsrc);

  //print_vector(n, mxGetPr(src));

  // Copy to GPU
  cudaMalloc((void **)&dst, svec);
  cudaMemcpy(dst, fsrc, svec, cudaMemcpyHostToDevice);

  free(fsrc);

  return dst;
}

mxArray *matlab_vector_from_gpu(int n, const double *src)
{
  size_t dims[2], svec;
  double *fdst;
  mxArray *dst;

  svec = sizeof(double) * n;

  // Copy to CPU
  fdst = (double *)malloc(svec);
  cudaMemcpy(fdst, src, svec, cudaMemcpyDeviceToHost);

  // Copy into matlab
  dims[0] = (size_t)n;
  dims[1] = 1;
  dst = mxArrayFromDoubleArray(2, dims, fdst);

  free(fdst);

  return dst;
}

gpu_mat_t matlab_matrix_to_gpu(const mxArray *src, bool transpose)
{
  gpu_mat_t dst;
  int *Ai, *Aj;
  double *Av;
  mwIndex *Ir, *Jc;
  double *Pr;

  int i, m, n, nnz;

  int iswap, *piswap;

  assert(mxIsSparse(src));
  assert(mxGetNumberOfDimensions(src) == 2);

  // Get matrix properties
  dst.format = GPU_MAT_CSC;
  m = dst.m = mxGetM(src);
  n = dst.n = mxGetN(src);

  // Retrieve matrix indices
  Ir = mxGetIr(src);
  Jc = mxGetJc(src);

  dst.nnz = nnz = Jc[n];

  // Copy to integer arrays
  Ai = (int *)malloc((nnz + n + 1) * sizeof(int) + nnz * sizeof(double));
  Aj = Ai + nnz;
  Av = (double *)(Aj + n + 1);

  for(i = 0; i < nnz; i++) {
    Ai[i] = (int)(Ir[i]);
  }
  for(i = 0; i < n + 1; i++) {
    Aj[i] = (int)(Jc[i]);
  }

  // Copy data
  assert(mxGetClassID(src) == mxDOUBLE_CLASS);
  Pr = mxGetPr(src);
  for(i = 0; i < nnz; i++) {
    Av[i] = (double)(Pr[i]);
  }

  // Copy to GPU
  cudaMalloc((void **)(&dst.i), nnz * sizeof(int));
  cudaMalloc((void **)(&dst.j), (n+1) * sizeof(int));
  cudaMalloc((void **)(&dst.v), nnz * sizeof(double));

  cudaMemcpy(dst.i, Ai, nnz * sizeof(int), cudaMemcpyHostToDevice);
  cudaMemcpy(dst.j, Aj, (n + 1) * sizeof(int), cudaMemcpyHostToDevice);
  cudaMemcpy(dst.v, Av, nnz * sizeof(double), cudaMemcpyHostToDevice);

  if(transpose) {
    dst.format = GPU_MAT_CSR;

    iswap = dst.m;
    dst.m = dst.n;
    dst.n = iswap;

    piswap = dst.i;
    dst.i = dst.j;
    dst.j = piswap;
  }

  return dst;
}

gpu_boxqp_t boxqp_gpu_from_matlab(const mxArray *Q, const mxArray *c, const mxArray *l, const mxArray *u)
{
  gpu_boxqp_t boxqp;

  boxqp.Q = matlab_matrix_to_gpu(Q, true);
  boxqp.c = matlab_vector_to_gpu(c);
  boxqp.l = matlab_vector_to_gpu(l);
  boxqp.u = matlab_vector_to_gpu(u);

  return boxqp;
}

void boxqp_gpu_cleanup(gpu_boxqp_t *boxqp)
{
  cudaFree(boxqp->Q.i);
  cudaFree(boxqp->Q.j);
  cudaFree(boxqp->Q.v);
  cudaFree(boxqp->c);
  cudaFree(boxqp->l);
  cudaFree(boxqp->u);
}

static const char * const gpfr_fields[] = {
  "grad_norm",
  "max_iters",

  "eta1",
  "eta2",

  "obj_const", 
  "print_interval"
};

static const int ngpfr_fields = sizeof(gpfr_fields) / sizeof(const char * const);

static void matlab_opts_printf(const char *s)
{
  mexPrintf("%s", s);
}

gpfr_options_t matlab_struct_to_gpfr_opts(const mxArray *matopts)
{
  gpfr_options_t opts;
  const mxArray *fields[ngpfr_fields];

  opts = gpfr_default_options;

  if(mxGetNumberOfElements(matopts) == 0) {
    return opts;
  }

  if(!mxIsStruct(matopts)) {
    mex_err("Expected options struct or []");
  }

  mxStructUnpack(ngpfr_fields, gpfr_fields, matopts, fields);

  if(fields[0] != NULL) {
    opts.grad_norm = mxArrayToDoubleScalar(fields[0]);
  }
  if(fields[1] != NULL) {
    opts.max_iters = mxArrayToIntScalar(fields[1]);
  }
  if(fields[2] != NULL) {
    opts.eta1 = mxArrayToDoubleScalar(fields[2]);
  }
  if(fields[3] != NULL) {
    opts.eta2 = mxArrayToDoubleScalar(fields[3]);
  }
  if(fields[4] != NULL) {
    opts.obj_const = mxArrayToIntScalar(fields[4]);
  }
  if(fields[5] != NULL) {
    opts.print_interval = mxArrayToIntScalar(fields[5]);
  }
  opts.printf = &matlab_opts_printf;

  return opts;
}

static const char *status_fields[] = {
  "objective",
  "grad_norm",

  "niters"
};

static int nstatus_fields = sizeof(status_fields) / sizeof(const char *);

mxArray *matlab_struct_from_optstatus(const boxopt_optstatus_t *s)
{
  mxArray *matst;
  mwSize dims[2];

  // Create options structure
  dims[0] = dims[1] = 1;
  matst = mxCreateStructArray(2, dims, nstatus_fields, status_fields);

  // Fill
  mxSetFieldByNumber(matst, 0, 0, mxArrayFromDoubleScalar(s->objective));
  mxSetFieldByNumber(matst, 0, 1, mxArrayFromDoubleScalar(s->grad_norm));
  mxSetFieldByNumber(matst, 0, 2, mxArrayFromIntScalar(s->niters));

  return matst;
}

/********************** Specific Functions **********************/

void matlab_boxqp(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
  cusparseHandle_t cusparse_handle;

  gpu_boxqp_t boxqp;
  int n;
  double *x;

  gpfr_options_t opts;
  boxopt_optstatus_t status;

  mex_assert(nlhs <= 2);
  mex_assert(nrhs == 6);

  mexPrintf("Starting CUDA libraries...\n");

  // Set up CUBLAS
  if(cublasInit() != CUBLAS_STATUS_SUCCESS) {
    mex_err("Could not start CUBLAS\n");
  }

  // Set up CUSPARSE
  if(cusparseCreate(&cusparse_handle) != CUSPARSE_STATUS_SUCCESS) {
    cublasShutdown();
     mex_err("Could not start CUSPARSE\n");
  }

  // Interpret arguments
  mexPrintf("Copying matrices...\n");
  boxqp = boxqp_gpu_from_matlab(prhs[0], prhs[1], prhs[2], prhs[3]);
  n = mxGetNumberOfElements(prhs[4]);
  x = matlab_vector_to_gpu(prhs[4]);
  opts = matlab_struct_to_gpfr_opts(prhs[5]);

  // Solve BoxQP
  mexPrintf("Starting gradient projection...\n");
  status = gpu_boxqp_matrix_grad_projf(x, &boxqp, &opts, cusparse_handle);

  // Return result 
  plhs[0] = matlab_vector_from_gpu(n, x);
  if(nlhs >= 2) {
    plhs[1] = matlab_struct_from_optstatus(&status);
  }

  // Cleanup
  cudaFree(x);

  boxqp_gpu_cleanup(&boxqp);

  cusparseDestroy(cusparse_handle);
  cublasShutdown();
}

///////////////////////////////// RWPrior /////////////////////////

typedef struct {
  gpu_mat_t L;
  gpu_mat_t H;
  gpu_mat_t HT;
  double *h;

  cusparseMatDescr_t desc;
  cusparseHandle_t cusparse_handle;
  struct cudaDeviceProp devprops;

} rwprior_data_t;

static void rwprior_Qmv(void *user_data, const double *x, double *Qx)
{
  rwprior_data_t *data;
  gpu_mat_t L, H, HT;
  double *h;
  int n, m;

  // Unpack arguments
  data = (rwprior_data_t *)user_data;
  L = data->L;
  H = data->H;
  HT = data->HT;
  h = data->h;

  assert(L.n == L.m);
  n = data->L.m;
  m = data->H.m;

  // Calculate 2 * (L * x + H' * (H * x)
  cusparseDcsrmv(data->cusparse_handle, CUSPARSE_OPERATION_NON_TRANSPOSE,
                 n, n,
                 1, data->desc, L.v, L.i, L.j, x,
                 0, Qx);

  cusparseDcsrmv(data->cusparse_handle, CUSPARSE_OPERATION_NON_TRANSPOSE,
                 m, n,
                 1, data->desc, H.v, H.i, H.j, x,
                 0, h);

  cusparseDcsrmv(data->cusparse_handle, CUSPARSE_OPERATION_NON_TRANSPOSE,
                 n, m,
                 1, data->desc, HT.v, HT.i, HT.j, h,
                 1, Qx);
  
  cublasDscal(n, 2, Qx, 1);
}

void matlab_rwprior(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
  cusparseHandle_t cusparse_handle;
  int device;
  rwprior_data_t rwprior;
  double *c, *l, *u, *x;
  int m, n;

  gpfr_options_t opts;
  boxopt_optstatus_t status;

  mex_assert(nlhs <= 2);
  mex_assert(nrhs == 7);

  // Set up CUBLAS
  if(cublasInit() != CUBLAS_STATUS_SUCCESS) {
    mex_err("Could not start CUBLAS\n");
  }

  // Set up CUSPARSE
  if(cusparseCreate(&cusparse_handle) != CUSPARSE_STATUS_SUCCESS) {
    cublasShutdown();
     mex_err("Could not start CUSPARSE\n");
  }
  rwprior.cusparse_handle = cusparse_handle;


  // Interpret arguments
  // [x status] = rwprior(L, H, c, l, u, x0, opts);
  rwprior.L = matlab_matrix_to_gpu(prhs[0], true);
  rwprior.HT = matlab_matrix_to_gpu(prhs[1], true);

  rwprior.H = gpu_mat_transpose(cusparse_handle, &rwprior.HT);
  c = matlab_vector_to_gpu(prhs[2]);
  l = matlab_vector_to_gpu(prhs[3]);
  u = matlab_vector_to_gpu(prhs[4]);
  x = matlab_vector_to_gpu(prhs[5]);
  opts = matlab_struct_to_gpfr_opts(prhs[6]);

  // Initialize rwprior structure
  m = rwprior.H.m;
  n = rwprior.H.n;

  cudaGetDevice(&device);
  cudaGetDeviceProperties(&rwprior.devprops, device);
  cusparseCreateMatDescr(&rwprior.desc);
  cudaMalloc((void **)&rwprior.h, m * sizeof(double));

  // Run GPFR
  status = gpu_boxqp_grad_proj(n, x, &rwprior, &rwprior_Qmv, c, l, u, &opts);

  // Return solution
  plhs[0] = matlab_vector_from_gpu(n, x);
  if(nlhs >= 2) {
    plhs[1] = matlab_struct_from_optstatus(&status);
  }

  double *Qx;
  cudaMalloc((void **)&Qx, n * sizeof(double));
  rwprior_Qmv(&rwprior, x, Qx);
  print_vector_gpu(4, Qx);
  cudaFree(Qx);

  // Cleanup
  cudaFree(rwprior.h);
  cudaFree(c);
  cudaFree(l);
  cudaFree(u);
  cudaFree(x);

  gpu_mat_free(&rwprior.L);
  gpu_mat_free(&rwprior.H);
  gpu_mat_free(&rwprior.HT);

  cusparseDestroyMatDescr(rwprior.desc);
  cusparseDestroy(cusparse_handle);
  cublasShutdown();
}

/////////////////////////////// RWCoseg /////////////////////////

typedef struct {
  int nimages;

  /**
   * The total number of pixels in all images.
   */
  int ntotal;

  gpu_mat_t* L;
  gpu_mat_t* H;
  gpu_mat_t* HT;
  gpu_mat_t Pinv;
  double **h;
  double lambda;
  double *Pinvx;

  cusparseMatDescr_t desc;
  cusparseHandle_t cusparse_handle;
  struct cudaDeviceProp devprops;

} rwcoseg_data_t;


static void matlab_print_vector(int n, const double *x)
{
  int i;

  mexPrintf("(");
  for(i = 0; i < n; i++) {
    if(i != 0) {
      mexPrintf(", ");
    }

    mexPrintf("%e", (float)(x[i]));
  }
  mexPrintf(")\n");
}

static void matlab_print_vector_gpu(int n, const double *x)
{
  double *cpu;

  cpu = malloc(n * sizeof(double));
  cudaMemcpy(cpu, x, n * sizeof(double), cudaMemcpyDeviceToHost);
  matlab_print_vector(n, cpu);
  free(cpu);
}


static void rwcoseg_Qmv(void *user_data, const double *x, double *Qx)
{
  rwcoseg_data_t *data;
  gpu_mat_t *L, *H, *HT, Pinv;
  double **h, lambda, *Pinvx;
  int n, m, i, k, nimages, ntotal;

  // Unpack arguments
  data = (rwcoseg_data_t *)user_data;
  nimages = data->nimages;
  Pinv = data->Pinv;
  Pinvx = data->Pinvx;
  L = data->L;
  H = data->H;
  HT = data->HT;
  h = data->h;
  lambda = data->lambda;

  // Precondition vector
  // Pinvx = Pinv * x
  cusparseDcsrmv(data->cusparse_handle, CUSPARSE_OPERATION_NON_TRANSPOSE,
                 Pinv.m, Pinv.n,
                 1, data->desc, Pinv.v, Pinv.i, Pinv.j, x,
                 0, Pinvx);

  k = 0;
  for(i = 0; i < nimages; i++) {
    assert(L[i].n == L[i].m);
    n = data->L[i].m;
    m = data->H[i].m;

    // Qx_i = L_i * Pinvx_i
    cusparseDcsrmv(data->cusparse_handle, CUSPARSE_OPERATION_NON_TRANSPOSE,
                   n, n,
                   1, data->desc, L[i].v, L[i].i, L[i].j, Pinvx + k,
                   0, Qx + k);

    // h_i = H_i * Pinvx_i
    cusparseDcsrmv(data->cusparse_handle, CUSPARSE_OPERATION_NON_TRANSPOSE,
                   m, n,
                   1, data->desc, H[i].v, H[i].i, H[i].j, Pinvx + k,
                   0, h[i]);

    // Qx_i = lambda * H_i' * h_i + Qx_i
    cusparseDcsrmv(data->cusparse_handle, CUSPARSE_OPERATION_NON_TRANSPOSE,
                   n, m,
                   lambda, data->desc, HT[i].v, HT[i].i, HT[i].j, h[i],
                   1, Qx + k);

    k += n;
  }
  ntotal = k;

  if(nimages != 2) {
    mex_err("Not currently handling %d images\n", nimages);
  }

  
  // Qx_0 <-- -lambda * H_0' * h_1 + Qx_0
  cusparseDcsrmv(data->cusparse_handle, CUSPARSE_OPERATION_NON_TRANSPOSE,
                 HT[0].m, HT[0].n,
                 -lambda, data->desc, HT[0].v, HT[0].i, HT[0].j, h[1],
                 1, Qx);

  // Qx_1 <-- -lambda * H_1' * h_0 + Qx_1
  k = L[0].n;
  cusparseDcsrmv(data->cusparse_handle, CUSPARSE_OPERATION_NON_TRANSPOSE,
                 HT[1].m, HT[1].n,
                 -lambda, data->desc, HT[1].v, HT[1].i, HT[1].j, h[0],
                 1, Qx + k);

  
  // Qx = 2 * Qx
  // To cancel out 1/2 in standard QP formulation
  cublasDscal(ntotal, 2, Qx, 1);
}

void matlab_rwcoseg(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
  cusparseHandle_t cusparse_handle;
  int device;
  rwcoseg_data_t rwcoseg;
  double *c, *l, *u, *x;
  int m, n, i, nimages;

  gpfr_options_t opts;
  boxopt_optstatus_t status;

  gpu_mat_t PinvT;

  // Set up CUBLAS
  if(cublasInit() != CUBLAS_STATUS_SUCCESS) {
    mex_err("Could not start CUBLAS\n");
  }

  // Set up CUSPARSE
  if(cusparseCreate(&cusparse_handle) != CUSPARSE_STATUS_SUCCESS) {
    cublasShutdown();
     mex_err("Could not start CUSPARSE\n");
  }
  rwcoseg.cusparse_handle = cusparse_handle;
  cusparseCreateMatDescr(&rwcoseg.desc);

  cudaGetDevice(&device);
  cudaGetDeviceProperties(&rwcoseg.devprops, device);

  ///////////// Interpret arguments ////////////////////
  // [x status] = rwprior(L, H, Pinv, l, u, lambda, x0, opts);
  mex_assert(nrhs == 8);

  // Read in L and H
  mex_assert(mxIsCell(prhs[0]) && mxIsCell(prhs[1]));
  nimages = rwcoseg.nimages = mxGetNumberOfElements(prhs[0]);
  mex_assert(mxGetNumberOfElements(prhs[1]) == nimages);

  rwcoseg.L = (gpu_mat_t *)malloc(sizeof(gpu_mat_t) * nimages * 3);
  rwcoseg.HT = rwcoseg.L + nimages;
  rwcoseg.H = rwcoseg.HT + nimages;
  rwcoseg.h = (double **)malloc(sizeof(double *) * nimages);

  rwcoseg.ntotal = 0;
  for(i = 0; i < nimages; i++) {
    rwcoseg.L[i] = matlab_matrix_to_gpu(mxGetCell(prhs[0], i), true);
    rwcoseg.HT[i] = matlab_matrix_to_gpu(mxGetCell(prhs[1], i), true);
    rwcoseg.H[i] = gpu_mat_transpose(cusparse_handle, rwcoseg.HT + i);

    m = rwcoseg.H[i].m;
    n = rwcoseg.H[i].n;
    cudaMalloc((void **)(rwcoseg.h + i), m * sizeof(double));

    rwcoseg.ntotal += n;
  }

  // Read in P
  PinvT = matlab_matrix_to_gpu(prhs[2], true);
  rwcoseg.Pinv = gpu_mat_transpose(cusparse_handle, &PinvT);
  gpu_mat_free(&PinvT);

  // Get other arguments
  l = matlab_vector_to_gpu(prhs[3]);
  u = matlab_vector_to_gpu(prhs[4]);
  rwcoseg.lambda = mxArrayToDoubleScalar(prhs[5]);
  x = matlab_vector_to_gpu(prhs[6]);
  opts = matlab_struct_to_gpfr_opts(prhs[7]);

  // Allocate c as zero vector
  cudaMalloc((void **)&c, rwcoseg.ntotal * sizeof(double));
  cudaMemset(c, 0, rwcoseg.ntotal * sizeof(double));
  
  // Allocate space for preconditioned vector
  cudaMalloc((void **)&rwcoseg.Pinvx, rwcoseg.ntotal * sizeof(double));

  // Call solver
  status = gpu_boxqp_grad_proj(rwcoseg.ntotal, x, &rwcoseg, &rwcoseg_Qmv, c, l, u, &opts);

  //double *Qx;
  //cudaMalloc((void **)&Qx, rwcoseg.ntotal * sizeof(double));
  //rwcoseg_Qmv(&rwcoseg, x, Qx);

  // Return solution
  plhs[0] = matlab_vector_from_gpu(rwcoseg.ntotal, x);
  if(nlhs >= 2) {
    plhs[1] = matlab_struct_from_optstatus(&status);
  }

  // Cleanup
  //cudaFree(Qx);

  cudaFree(rwcoseg.Pinvx);
  cudaFree(c);
  cudaFree(x); cudaFree(u); cudaFree(l);

  gpu_mat_free(&rwcoseg.Pinv);

  for(i = 0; i < nimages; i++) {
    cudaFree(rwcoseg.h[i]);

    gpu_mat_free(&(rwcoseg.L[i]));
    gpu_mat_free(&(rwcoseg.H[i]));
    gpu_mat_free(&(rwcoseg.HT[i]));
  }
  free(rwcoseg.h);
  free(rwcoseg.L);

  cusparseDestroyMatDescr(rwcoseg.desc);
  cusparseDestroy(cusparse_handle);
  cublasShutdown();
}
