if __name__ != '__main__':
    from SCons.Script import *
import sys, os, os.path
import re

cpp_re = '^[ \t]*#[ \t]*(include|import|ifdef|ifndef|if|endif|define|pragma|warning|error)'
cpp_re = re.compile(cpp_re, re.IGNORECASE)

include_re = '^[ \t]*#[ \t]*(?:include|import)[ \t]*(<|")([^>"]+)(>|")'
include_re = re.compile(include_re)

ifndef_re = '^[ \t]*#[ \t]*ifndef[ \t]+([^\s]*)'
ifndef_re = re.compile(ifndef_re)

define_re = '^[ \t]*#[ \t]*define[ \t]+([^\s]*)'
define_re = re.compile(define_re)

def preprocessorCommand(line):
    m = cpp_re.match(line)
    if m == None:
        return None
    return m.groups()[0]

def includedFile(env, path, line):
    """Returns the full path of any #include directive in line."""

    m = include_re.match(line)
    if m == None:
        return None

    # Search local path
    def searchLocal(root, path):
        local = os.path.join(os.path.dirname(root), path)
        if os.path.exists(local):
            return local
        return None

    # Search CPPPATH
    def searchCPPPATH(path):
        for d in env['CPPPATH']:
            if type(d) == str:
                d = env.Dir(d)

            fullpath = os.path.join(d.abspath, path)
            if os.path.exists(fullpath):
                return fullpath
        return None

    # Extract full path to included file 
    fullpath = None
    if m.groups()[0] == '"':
        fullpath = searchLocal(path, m.groups()[1])
    if fullpath == None:
        fullpath = searchCPPPATH(m.groups()[1])

    return fullpath

def addHeader(out, path, env, included, root, include_guard_prefix = ''):
    iflevel = 0
    include_guard_iflevels = set()

    def isIncludeGuard(line):
        def prefixMatch(obj):
            return obj[:len(include_guard_prefix)] == include_guard_prefix

        m = ifndef_re.match(line)
        if m != None and prefixMatch(m.groups()[0]):
            return True

        m = define_re.match(line)
        if m != None and prefixMatch(m.groups()[0]):
            return True

        return False

    # Partial CPP simulation
    for i, line in enumerate(file(path, 'r')):
        cpp = preprocessorCommand(line)
        if not cpp:
            out.write(line)
            continue

        # Include files
        if cpp in ['include', 'import']:
            include = includedFile(env, path, line)
            if include == None:
                out.write(line)
                continue

            abspath = env.File(include).abspath

            # Filter out repeat includes
            if abspath in included:
                continue
            included.add(abspath)
            print abspath

            # Output system includes
            if abspath[:len(root)] != root:
                out.write(line)
                continue

            # Input file
            addHeader(out, abspath, env, included, root, include_guard_prefix)
            continue

        # Include guards
        if cpp == 'ifndef' and isIncludeGuard(line):
            include_guard_iflevels.add(iflevel)
            iflevel += 1
            continue

        if cpp == 'define' and isIncludeGuard(line):
            continue

        if cpp in ['if', 'ifdef']:
            iflevel += 1
        if cpp == 'endif':
            iflevel -= 1
            if iflevel in include_guard_iflevels:
                continue

        out.write(line)

def aggregate_header(target, source, env):
    # Retrieve arguments
    root = None
    headers = []
    for s in source:
        if s.__class__ == env.Dir('#').__class__:
            break
        else:
            headers.append(str(s))

    source = source[len(headers):]
    root = source[0].abspath
    include_guard_prefix = str(source[1])
    target = file(str(target[0]), 'w+')

    # Add each header in the source list
    included = set()
    for h in headers:
        addHeader(target, str(h), env, included, root, include_guard_prefix)

    return

def method(env, target, source, root = None, include_guard_prefix = ''):
    if type(source) != list:
        source = [source]

    if root == None:
        root = env.Dir('#')

    env.AggregateHeaderBuilder(target, source + [env.Dir(root), env.Value(include_guard_prefix)])

def generate(env):
    bld = env.Builder(action = aggregate_header, source_scanner = CScanner)
    env.Append(BUILDERS = {'AggregateHeaderBuilder': bld})
    env.AddMethod(method, 'AggregateHeader')

def exists(env):
    return True

if __name__ == '__main__':
    class FakeEnv:
        class Dir:
            def __init__(self, s):
                self.s = s
                self.abspath = os.path.abspath(s)

            def __str__(self):
                return self.s

        File = Dir

    env = FakeEnv()
    sources = sys.argv[1:-1]
    sources += [env.Dir('.'), sys.argv[-1]]
    aggregate_header(['/dev/stdout'], sources, env)
