import os, os.path, subprocess
import re
from SCons.Script import *

SVNDEBUG = True

class SVNError:
    def __init__(self, returncode, stderr = None):
        self.returncode = returncode
        self.stderr = stderr

    def __str__(self):
        s = 'SVN Error %d' % self.returncode
        if self.stderr:
            s += '\n' + self.stderr
        return s

class SVNInfo:
    """Class to retrieve and store the SVN info for a given WC path or URL."""

    entry_re = re.compile('^([^\:]+)\:\s*(.*?)\s*$')

    def __init__(self, path_or_uri, svn_command = 'svn'):
        # Build command
        cmd = [svn_command, 'info', path_or_uri]

        if SVNDEBUG:
            print ' '.join(cmd)
        proc = subprocess.Popen(cmd, stdout = subprocess.PIPE)

        # Parse output
        for line in proc.stdout:
            m = self.entry_re.match(line)
            if m == None:
                continue

            field, val = m.groups()
            if field == 'Path':
                self.path = val
            elif field == 'URL':
                self.url = val
            elif field == 'Repository Root':
                self.repo_root = val
            elif field == 'Repository UUID':
                self.repo_uuid = val
            elif field == 'Revision':
                self.rev = int(val)
            elif field == 'Last Changed Rev':
                self.changed_rev = int(val)
            elif field == 'Node Kind':
                self.node_kind = val

        returncode = proc.wait()
        if returncode != 0:
            raise SVNError(returncode)

    def __str__(self):
        s  = 'Path: %s\n' % self.path
        s += 'URL: %s\n' % self.url
        s += 'Repository Root: %s\n' % self.repo_root
        s += 'Repository UUID: %s\n' % self.repo_uuid
        s += 'Revision: %s\n' % self.rev
        s += 'Last Changed Rev: %s\n' % self.changed_rev

        return s

    def __repr__(self):
        return str(self)


    head_revs = {}

    def headRevision(self):
        """Get the head revision # for this object's path or URI."""
        if self.repo_uuid not in self.head_revs:
            self.head_revs[self.repo_uuid] = SVNInfo(self.repo_root).rev
            
        return self.head_revs[self.repo_uuid]

    server_re = re.compile('(?:http|svn)\:\/\/[^\/]+\/')

    def server(self):
        """Get the server URL (ie 'svn://server/') of the repository."""
        
        m = self.server_re.match(self.repo_root)
        if m == None:
            return None

        return m.group()

def SVNlist(path_or_uri,
            recurse = False, max_depth = None,
            svn_command = 'svn'):
    """List the contents of a directory in SVN."""

    # Construct command
    cmd = [svn_command, 'ls']
    if recurse:
        cmd += ['-R']
    if max_depth != None:
        cmd += ['--depth', str(max_depth)]
    cmd += [path_or_uri]

    if SVNDEBUG:
        print ' '.join(cmd)
    proc = subprocess.Popen(cmd, stdout = subprocess.PIPE)

    # Parse the output
    for line in proc.stdout:
        line = line.strip()
        if line == '':
            continue
        yield line

def SVNisURL(s):
    """Check whether the given string corresponds to an SVN URL"""

    svn_url_re = '[a-z]+\:\/\/'
    svn_url_re = re.compile(svn_url_re, re.IGNORECASE)

    return svn_url_re.match(s) != None
        

def generator(source, target, env, for_signature):
    """Generates the subversion command issued by SCons."""
    url, path, revision = source[0].value

    try: # TODO: better way of figuring out if we have version control...
        SVNInfo(path)

        # Run update command
        # TODO: compare url, delete and re-checkout if bad
        return '$SVN update -r%d %s' % (revision, path)
    except:
        # Run checkout command
        return [Delete(path),
                '$SVN checkout -r%d %s %s' % (revision, url, path)]

def emit(target, source, env):
    """Modify the target list to include all files produced by SVN checkout."""

    assert(len(target) == 1 and len(source) == 1)

    root_url, root_path, rev = source[0].value
    
    # Get target list
    for path in SVNlist(root_path,
                        recurse = True, max_depth = None,
                        svn_command = env['SVN']):
        path = os.path.join(root_path, path)

        if path[-1] == '/':
            # Directory
            pass
        else:
            # File
            target.append(env.File(path))

    return target, source

def update(path, url, revision = None, svn_command = 'svn'):
    """Directly call the SVN update/checkout command through python."""

    try:
        current_revision = SVNInfo(path).rev
    except:
        current_revision = 0

    # Don't update if up-to-date
    if revision and current_revision >= revision:
        return

    subprocess.call('pwd')

    # Determine command
    if current_revision == 0:
        Execute(Delete(path))
        cmd = [svn_command, 'checkout']
        if revision:
            cmd += ['-r%d' % revision]
        cmd += [url, path]
    else:
        cmd = [svn_command, 'update']
        if revision:
            cmd += ['-r%d' % revision]
        cmd += [path]

    # Run SVN
    if SVNDEBUG:
        print ' '.join(cmd)
    subprocess.call(cmd)

def subversion(env, path, url, revision = None):
    """SCons environment method which wraps the builder."""

    # Normalize input
    if not SVNisURL(url):
        url = env['SVNSERVE'] + url
    if not revision:
        try:
            revision = SVNInfo(url).headRevision()
        except:
            revision = None

    if type(path) == list:
        for p, u in zip(path, url):
            subversion(env, p, u, revision = revision)
        return

    if type(path) != str:
        path = path.abspath

    # Call builder
    #files = env.SVNBuilder(path, env.Value((url, path.path, revision)))
    #env.Precious(files)
    #env.NoClean(files)

    # Directly update
    update(path, url, revision)

    #return files

def generate(env):
    """Set up environment for SVN builder."""

    env['SVN'] = 'svn'
    try:
        root = env.Dir('#').abspath
        env['SVNSERVE'] = SVNInfo(root).server()
    except:
        env['SVNSERVE'] = 'svn://localhost/' # TODO: get from SVN info of #

    # Create builder
    bld = env.Builder(generator = generator, emitter = emit)
    env.Append(BUILDERS = {'SVNBuilder': bld})

    # Add method
    env.AddMethod(subversion, 'SVN')

def exists(env):
    return env.Detect('svn')

#print SVNInfo('svn://localhost/research/util/toolbox').headRevision()
#print SVNInfo('svn://localhost/research/QAP/trunk/QAP').headRevision()
#print SVNInfo('svn://localhost/').headRevision()
#print SVNInfo('svn://dijkstra/research/QAP/trunk/QAP').server()

#for entry in SVNlist('svn://localhost/research/curves/trunk', recurse = True):
#    print entry
