#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <stdint.h>
#include <assert.h>

#include <sys/time.h>

#include <cuda_runtime.h>
#include <cublas.h>
#include <cusparse.h>

#include "boxqp.h"
#include "debug.h"

/////////////////////////// Matrix Operations /////////////////////

gpu_mat_t gpu_mat_transpose(cusparseHandle_t handle, const gpu_mat_t *A)
{
  int m, n, nnz;
  gpu_mat_t T;
  cusparseStatus_t s;

  // Set sizes and format
  T.format = GPU_MAT_CSR;
  m = T.m = A->n;
  n = T.n = A->m;
  nnz = T.nnz= A->nnz;

  // Allocate
  T.i = T.j = NULL;
  T.v = NULL;
  cudaMalloc((void **)&T.i, sizeof(int) * (m + 1));
  cudaMalloc((void **)&T.j, sizeof(int) * nnz);
  cudaMalloc((void **)&T.v, sizeof(double) * nnz);

  if(T.i == NULL || T.j == NULL || T.v == NULL) {
    fprintf(stderr, "Allocation error\n");
  }

  // Convert indices and copy data
  s = cusparseDcsr2csc(
    handle,
    n, m,
    A->v, A->i, A->j,
    T.v,  T.j,  T.i,
    1, CUSPARSE_INDEX_BASE_ZERO); 
  if(s != CUSPARSE_STATUS_SUCCESS) {
    fprintf(stderr, "csr2csc error %d", (int)s);
  }

  return T;
}

void gpu_mat_free(gpu_mat_t *mat)
{
  cudaFree(mat->i);
  cudaFree(mat->j);
  cudaFree(mat->v);
}

////////////////////////// QP closure ///////////////////////////

typedef struct {
  int n;
  void *user_data;
  void (*Qmv)(void *user_data, const double *x, double *Qx);
  const double *c;

  double *work;

} qmv_calc_t;

static double qp_objective(void *user_data, int n, const double *x)
{
  qmv_calc_t *calc;
  const gpu_boxqp_t *QP;
  double *Qx, obj;

  // Unpack arguments
  calc = (qmv_calc_t *)user_data;
  assert(n == calc->n);

  // Calculate 1/2 * x' * Q * x
  Qx = calc->work;
  calc->Qmv(calc->user_data, x, Qx);
  obj = cublasDdot(n, Qx, 1, x, 1)/2;

  // Add in c' * x
  obj += cublasDdot(n, calc->c, 1, x, 1);

  return obj;
}

static void qp_gradient(void *user_data, int n, const double *x, double *g)
{
  qmv_calc_t *calc;
  const gpu_boxqp_t *QP;
  
  // Unpack arguments
  calc = (qmv_calc_t *)user_data;
  assert(n == calc->n);

  // Calculate Q * x + c
  calc->Qmv(calc->user_data, x, g);
  cublasDaxpy(n, 1, calc->c, 1, g, 1);
}

static double qp_directional_derivative2(void *user_data, int n, const double *x, const double *d)
{
  qmv_calc_t *calc;
  const gpu_boxqp_t *QP;
  double *Qd;

  // Unpack arguments
  calc = (qmv_calc_t *)user_data;
  assert(n == calc->n);

  // Calculate d * Q * d
  Qd = calc->work;
  calc->Qmv(calc->user_data, d, Qd);
  return cublasDdot(n, d, 1, Qd, 1);
}


boxopt_optstatus_t gpu_boxqp_grad_proj(int n, double *x, 
  void *user_data, void (*Qmv)(void *user_data, const double *x, double *Qx),
  const double *c, const double *l, const double *u,
  const gpfr_options_t *opts)
{
  qmv_calc_t calc;
  gpu_functionf_t f;
  boxopt_optstatus_t status;

  // Initialize data
  calc.n = n;
  calc.user_data = user_data;
  calc.Qmv = Qmv;
  calc.c = c;
  cudaMalloc((void **)&calc.work, n * sizeof(double));

  // Fill in objective
  f.objective = qp_objective;
  f.gradient = qp_gradient;
  f.directional_derivative2 = qp_directional_derivative2;

  f.user_data = &calc;

  // Call gradient projection algorithm
  status = boxopt_gpu_gpfr(n, x, l, u, &f, opts);

  // Cleanup
  cudaFree(calc.work);

  return status;
}

typedef struct {
  gpu_mat_t Q;

  cusparseMatDescr_t desc;
  cusparseHandle_t cusparse_handle;

  struct cudaDeviceProp devprops;


} Qdata_t;

#include "mv.h"

// Calculate Q*x for sparse CSR Q
static void Qmv(void *user_data, const double *x, double *Qx)
{
  Qdata_t *data;

  cusparseStatus_t s;
  gpu_mat_t Q;
  int n;
  cudaError_t err;

  data = (Qdata_t *)user_data;

  // Unpack arguments
  Q = data->Q;
  n = Q.n;

  assert(Q.m == Q.n);
  assert(Q.format == GPU_MAT_CSR);

  //printf("Qmv\n");
  //print_vector_gpu(n, x);

  // Call CUSPARSE multiplication
  s = cusparseDcsrmv(data->cusparse_handle, CUSPARSE_OPERATION_NON_TRANSPOSE,
                     n, n,
                     1, data->desc, Q.v, Q.i, Q.j, x,
                     0, Qx);
  cudaThreadSynchronize();
  if(s != CUSPARSE_STATUS_SUCCESS) {
    printf("Qmv failed, status = %d\n", (int)s);
  }

  //print_vector_gpu(n, Qx);

  //err = cudaGetLastError();
  //printf("%s\n", cudaGetErrorString(err));

}


boxopt_optstatus_t gpu_boxqp_matrix_grad_projf(double *x, const gpu_boxqp_t *QP, const gpfr_options_t *opts, cusparseHandle_t cusparse_handle)
{
  Qdata_t Qdata;
  int device;
  boxopt_optstatus_t status;

  // Initialize data
  Qdata.Q = QP->Q;

  cusparseCreateMatDescr(&Qdata.desc);
  Qdata.cusparse_handle = cusparse_handle;
  cudaGetDevice(&device);
  cudaGetDeviceProperties(&Qdata.devprops, device);

  // Call base function
  status = gpu_boxqp_grad_proj(QP->Q.n, x, &Qdata, &Qmv, QP->c, QP->l, QP->u, opts);

  // Cleanup
  cusparseDestroyMatDescr(Qdata.desc);

  return status;
}
