#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <sys/time.h>
#include <stdint.h>
#include <stdarg.h>

#include <cuda_runtime.h>
#include <cublas.h>

#include "grad_proj.h"
#include "gpfr.h"

#include "norm.h"
#include "project.h"
#include "set.h"
#include "runparams.h"
#include "debug.h"

static void default_print_function(const char *s)
{
  fputs(s, stdout);
}

const grad_proj_options_t grad_proj_default_options = {
  .grad_norm = 1e-6,
  .max_iters = 0
};

const gpfr_options_t gpfr_default_options = {
  .grad_norm = 1e-6,
  .max_iters = 0,

  .eta1 = 1e-2,
  .eta2 = 1e-2,

  .obj_const = 0,
  .print_interval = 1,
  .printf = default_print_function
};

typedef struct {
  int n;

  const double *l, *u;
  const gpu_functionf_t *f;

  
} prob_t;

#ifndef INFINITY
#define INFINITY (1.0/0.0)
#endif

static void print(void (*print_function)(const char *), const char *format, ...)
{
  char s[1024];
	va_list args;

  // Format string
	va_start(args, format);
	vsnprintf(s, 1023, format, args);
	va_end(args);

  // Output to function
  print_function(s);
}

/////////////////////////////////// Line Search //////////////////////////

/**
 * Evaluate the objective along a line.
 */
static double phi(int n, const gpu_functionf_t *f, double *xa, const double *x, const double *d, double alpha, const double *l, const double *u, struct cudaDeviceProp *devprops)
{
  // Find line point
  cudaMemcpy(xa, x, n * sizeof(double), cudaMemcpyDeviceToDevice);
  cublasDaxpy(n, alpha, d, 1, xa, 1);

  // Project onto box
  boxproject_gpu(n, xa, l, u, devprops);

  // Evaluate objective
  return f->objective(f->user_data, n, xa);
}

static double mid(double a, double b, double c)
{
  if(a > b) {
    return mid(b, a, c);
  }
  if(b > c) {
    return mid(a, c, b);
  }

  return b;
}

static double line_search(int n, const gpu_functionf_t *f, double *xa, const double *x, const double *g, const double *d, const double *l, const double *u, double *b, struct cudaDeviceProp *devprops)
{
  double alpha, dphi0, ddphi0, beta, phi0, phialpha;
  double a, alphastar;

  beta = boxproject_breakpoint_gpu(n, x, d, l, u, b, devprops);
  //printf("beta = %e\n", beta);

  // Get initial estimate
  dphi0 = cublasDdot(n, g, 1, d, 1);
  ddphi0 = f->directional_derivative2(f->user_data, n, x, d);
  //printf("dphi = %e, ddphi0 = %e\n", dphi0, ddphi0);

  if(ddphi0 == 0) {
    alpha = beta;
  }
  else {
    alpha = -dphi0/ddphi0;
  }

  // If our "exact" line search stops before the breakpoint,
  // we will already have sufficient decrease.
  if(alpha <= beta) {
    cudaMemcpy(xa, x, n * sizeof(double), cudaMemcpyDeviceToDevice);
    cublasDaxpy(n, alpha, d, 1, xa, 1);

    //printf("alpha < beta = %e\n", alpha);

    return alpha;
  }

  // Otherwise, do backtracking line search...
  int i;
  for(i = 0; i < 10; i++) {

    // Check sufficient decrease
    phi0 = f->objective(f->user_data, n, x);
    phialpha = phi(n, f, xa, x, d, alpha, l, u, devprops);

    cudaMemcpy(b, xa, n * sizeof(double), cudaMemcpyDeviceToDevice);
    cublasDaxpy(n, -1, x, 1, b, 1); // b = xa - x

    /*
    printf("alpha = %e\n", alpha);
    printf("distance = %e, decrease = %e\n", linfty_normf_gpu(n, b), phialpha - phi0);
    printf("compare = %e\n", (phialpha - phi0) / cublasDdot(n, g, 1, b, 1));
    */
    if((phialpha - phi0) / cublasDdot(n, g, 1, b, 1) > 0.5) {
      break;
    }

    // Backtrack
    a = (phialpha - dphi0 * alpha - phi0) / (alpha * alpha);

    alphastar = -dphi0 / (2 * a);
    alpha = mid(alpha / 100, alphastar, alpha * 2.0/3);
  }

  //printf("alphasearch = %e\n", alpha);

  return alpha;
}


//////////////////////////// Gradient Projection /////////////////////

static void projected_gradient_direction(
  int n, double *g, double *d, const double *x, const double *l, const double *u,
  const gpu_functionf_t *f, struct cudaDeviceProp *devprops)
{
  // Find gradient
  f->gradient(f->user_data, n, x, g);

  // Take negative
  cudaMemcpy(d, g, n * sizeof(double), cudaMemcpyDeviceToDevice);
  cublasDscal(n, -1, d, 1);

  // Project direction
  boxproject_direction_gpu(n, d, x, l, u, devprops);
}

boxopt_optstatus_t gpu_grad_projf(
  int n, double *param_x, const double *l, const double *u,
  const gpu_functionf_t *f, const grad_proj_options_t *opts)
{
  double *x, *xa, *xswap, *g, *d, *b;
  size_t svec;

  int device;
  struct cudaDeviceProp devprops;

  int iter;
  double last_objval, objval, dnorm;

  boxopt_optstatus_t status;

  if(opts == NULL) {
    opts = &grad_proj_default_options;
  }

  printf("Beginning projected gradient descent...\n");

  // Get device properties
  cudaGetDevice(&device);
  cudaGetDeviceProperties(&devprops, device);

  // Allocate vectors
  svec = sizeof(double) * n;

  x = param_x;
  cudaMalloc(&xa, 4 * svec);
  g = xa + n;
  d = g + n;
  b = d + n;

  boxproject_gpu(n, x, l, u, &devprops);

  objval = f->objective(f->user_data, n, x);

  for(iter = 1; opts->max_iters <= 0 || iter <= opts->max_iters; iter++) {
    // Generate next direction
    projected_gradient_direction(n, g, d, x, l, u, f, &devprops);


    // Check stopping conditions
    dnorm = linfty_normf_gpu(n, d);

    printf("%4d %e %e %e\n", iter, objval, dnorm, last_objval - objval);
    if(dnorm < opts->grad_norm) {
      break;
    }

    if(objval > last_objval) {
      printf("Failing to minimize\n");
    }

    // Perform projected line search
    line_search(n, f, xa, x, g, d, l, u, b, &devprops);
    xswap = x;
    x = xa;
    xa = xswap;

    last_objval = objval;
    objval = f->objective(f->user_data, n, x);
  }

  // Copy result to output
  if(x != param_x) {
    cudaMemcpy(param_x, x, svec, cudaMemcpyDeviceToDevice);
    cudaFree(x);
  }
  else {
    cudaFree(xa);
  }

  status.objective = objval;
  status.grad_norm = dnorm;
  status.niters = iter-1;


  return status;
}

//////////////////////////// GPFR /////////////////////

static void projected_conjugate_gradient_direction(
  int n, double *g, double *d, double *lastg,
  const double *x, const double *l, const double *u,
  const gpu_functionf_t *f, struct cudaDeviceProp *devprops)
{
  double beta;

  // Get previous direction and gradient, projected into current plane
  cublasDscal(n, -1, lastg, 1);
  boxproject_direction_gpu(n, lastg, x, l, u, devprops);
  cublasDscal(n, -1, lastg, 1);

  // Find current gradient
  f->gradient(f->user_data, n, x, g);

  // TODO: better gradient projection
  cublasDscal(n, -1, g, 1);
  boxproject_direction_gpu(n, g, x, l, u, devprops);
  cublasDscal(n, -1, g, 1);

  // Find conjugacy constant
  beta = l2_normf_gpu(n, g) / l2_normf_gpu(n, lastg);
  beta = beta * beta;

  // Set d = P[-g + beta * d]
  cublasDscal(n, beta, d, 1);
  cublasDaxpy(n, -1, g, 1, d, 1);
  boxproject_direction_gpu(n, d, x, l, u, devprops);
}


static __global__ void active_set_kernel(
  int n, bool *aset, bool *bset,
  const double *g, const double *x, const double *l, const double *u)
{
  int i;

  // Get index
  i = blockDim.x * blockIdx.x + threadIdx.x;
  if(i >= n) {
    return;
  }

  aset[i] = false;
  bset[i] = false;

  if(x[i] <= l[i]) {
    aset[i] = true;
    
    if(g[i] < 0) {
      bset[i] = true;
    }
  }
  else if(x[i] >= u[i]) {
    aset[i] = true;

    if(g[i] > 0) {
      bset[i] = true;
    }
  }
}

void active_set(
  int n, bool *aset, bool *bset,
  const double *g, const double *x, const double *l, const double *u,
  const struct cudaDeviceProp *devprops)
{
	runparams_t p;

	p = runparams(n, devprops);

  while(n > 0) {
    active_set_kernel<<<p.blocks, p.threads>>>(n, aset, bset, g, x, l, u);

    aset += p.stride;
    bset += p.stride;

    g += p.stride;
    x += p.stride;
    l += p.stride;
    u += p.stride;

    n -= p.stride;
  }
}

#define MODE_PG (1)
#define MODE_FR (2)

static inline const char *mode_string(int mode)
{
  switch(mode) {
  case MODE_PG:
    return "PG";
  case MODE_FR:
    return "FR";
  default:
    return "Unknown mode";
  }
}

boxopt_optstatus_t boxopt_gpu_gpfr(
  int n, double *param_x, const double *l, const double *u,
  const gpu_functionf_t *f, const gpfr_options_t *opts)
{
  double *x, *xa, *fswap, *g, *d, *lastg, *b;
  size_t svec;

  bool *aset, *bset, *last_aset, *seteq_work, *seteq_host_work, *bswap;

  int device;
  struct cudaDeviceProp devprops;

  int iter;
  double last_objval, objval, gnorm, best_change, alpha;

  int mode;

  boxopt_optstatus_t status;

  if(opts == NULL) {
    opts = &gpfr_default_options;
  }
  // Initialize status
  status.niters = 0;

  printf("Beginning projected gradient descent...\n");

  // Get device properties
  cudaGetDevice(&device);
  cudaGetDeviceProperties(&devprops, device);

  // Allocate vectors
  svec = sizeof(double) * n;

  //printf("x0 = "); print_vector_gpu(n, param_x);

  x = param_x;
  cudaMalloc((void **)&xa, 5 * svec);
  g = xa + n;
  d = g + n;
  lastg = d + n;
  b = lastg + n;

  if(xa == NULL) {
    print(opts->printf, "Allocation error\n");
    return status;
  }

  cudaMalloc((void **)&aset, (6 * n) * sizeof(bool));
  last_aset = aset + n;
  bset = last_aset + n;
  seteq_work = bset + n; // length 3n

  if(aset == NULL) {
    print(opts->printf, "Allocation error\n");
    cudaFree(xa);
    return status;
  }

  seteq_host_work = (bool *)malloc(n * sizeof(bool));
  if(seteq_host_work == NULL) {
    print(opts->printf, "Allocation error\n");
    cudaFree(xa);
    cudaFree(aset);
    return status;
  }

  // Initial state
  mode = MODE_PG;
  best_change = -INFINITY;
  boxproject_gpu(n, x, l, u, &devprops);
  objval = f->objective(f->user_data, n, x);

  // Iterations
  for(iter = 1; opts->max_iters <= 0 || iter <= opts->max_iters; iter++) {
    // Generate next direction
    if(mode == MODE_FR) {
      projected_conjugate_gradient_direction(n, g, d, lastg, x, l, u, f, &devprops);
      gnorm = linfty_normf_gpu(n, g);
    }
    else {
      projected_gradient_direction(n, g, d, x, l, u, f, &devprops);
      gnorm = linfty_normf_gpu(n, d);
    }

    if(opts->print_interval > 0 && iter % opts->print_interval == 0) {
      print(opts->printf, "%4d %s % e % e % e\n",
            iter, mode_string(mode), objval + opts->obj_const, gnorm,
            last_objval - objval);
    }
    if(gnorm < opts->grad_norm) {
      break;
    }

    if(objval > last_objval) {
      print(opts->printf, "Failing to minimize\n");
    }

    // Perform projected line search
    alpha = line_search(n, f, xa, x, g, d, l, u, b, &devprops);
    //printf("alpha = % e\n", alpha);

    // Check mode switch conditions
    active_set(n, aset, bset, g, x, l, u, &devprops);
    if(last_objval - objval > best_change) {
      best_change = last_objval - objval;
    }

    if(mode == MODE_PG) {
      if(iter != 1 && sets_equal(n, aset, last_aset, seteq_work, seteq_host_work, &devprops)
         || last_objval - objval < opts->eta2 * best_change) {

        mode = MODE_FR;
        best_change = -INFINITY;
      }
    }
    else {
      // TODO: broken
      if(!sets_equal(n, aset, bset, seteq_work, seteq_host_work, &devprops)
         || last_objval - objval < opts->eta1 * best_change) {

        mode = MODE_PG;
        best_change = -INFINITY;
      }
    }

    // Swap vectors
    fswap = x;
    x = xa;
    xa = fswap;

    fswap = lastg;
    lastg = g;
    g = fswap;

    bswap = last_aset;
    last_aset = aset;
    aset = bswap;

    last_objval = objval;
    objval = f->objective(f->user_data, n, x);
  }

  // Copy result to output
  if(x != param_x) {
    cudaMemcpy(param_x, x, svec, cudaMemcpyDeviceToDevice);
    cudaFree(x);
  }
  else {
    cudaFree(xa);
  }

  status.objective = objval;
  status.grad_norm = gnorm;
  status.niters = iter-1;
  return status;
}
