#include "../matrix_internal.h"
#include "gpu.h"

#include <assert.h>
#include <stdlib.h>

#include <cuda_runtime.h>

#include "type_defs.h"

typedef struct gpu_sparse_matrix gpu_sparse_t;
struct gpu_sparse_matrix
{
	matrix_t base;

	cusparseHandle_t handle;
	cusparseMatDescr_t desc;

	int nnz;

	// Data (on GPU)
	int *i, *j;
	ELT *v;
};

static inline gpu_sparse_t *cast(matrix_t *base)
{
	assert(base->gpu && matrix_storage_is_sparse(matrix_storage_format(base)));
	return (gpu_sparse_t *)base;
}

static inline const gpu_sparse_t *cast_const(const matrix_t *base)
{
	assert(base->gpu && matrix_storage_is_sparse(matrix_storage_format(base)));
	return (const gpu_sparse_t *)base;
}

static const matrix_vtable_t csr_vtable;
static const matrix_vtable_t csc_vtable;
static const matrix_vtable_t coo_vtable;

////////////////////////////// Constructors /////////////////////

static void init_desc(gpu_sparse_t *A)
{
}

matrix_t *matrix_gpu_create_csr(cusparseHandle_t cusparse, linalg_elt_type_t type, int m, int n, int nnz)
{
	gpu_sparse_t *A;
	size_t s;

	A = (gpu_sparse_t *)malloc(sizeof(gpu_sparse_t) + 2 * sizeof(int));
  if(A == NULL) {
    return NULL;
  }

  // Initialize base class
  A->base.v = &csr_vtable;
  A->base.gpu = true;
  A->base.format = MATRIX_CSR;
  A->base.type = ELT_TYPE;
  
  A->base.ndims = 2;
  A->base.dims = (int *)(A + 1);
  A->base.dims[0] = m;
  A->base.dims[1] = n;

  // Initialize GPU matrix properties
  A->handle = cusparse;
  cusparseCreateMatDescr(&A->desc);

	A->nnz = nnz;

	// Allocate data on GPU
	s = sizeof(int) * (m + 1 + nnz) + sizeof(ELT) * nnz;
	cudaMalloc((void **)&A->i, s);
	if(A->i == NULL) {
		// Deinit base class
		free(A);
		return NULL;
	}
	A->j = A->i + m + 1;
	A->v = (ELT *)(A->j + nnz);

	return (matrix_t *)A;
}

matrix_t *matrix_gpu_create_csc(cusparseHandle_t cusparse, linalg_elt_type_t type, int m, int n, int nnz)
{
	gpu_sparse_t *A;
	size_t s;

	A = (gpu_sparse_t *)malloc(sizeof(gpu_sparse_t) + 2 * sizeof(int));
  if(A == NULL) {
    return NULL;
  }

  // Initialize base class
  A->base.v = &csc_vtable;
  A->base.gpu = true;
  A->base.format = MATRIX_CSC;
  A->base.type = ELT_TYPE;
  
  A->base.ndims = 2;
  A->base.dims = (int *)(A + 1);
  A->base.dims[0] = m;
  A->base.dims[1] = n;

  // Initialize GPU matrix properties
  A->handle = cusparse;
  cusparseCreateMatDescr(&A->desc);

	A->nnz = nnz;

	// Allocate data on GPU
	s = sizeof(int) * (n + 1 + nnz) + sizeof(ELT) * nnz;
	cudaMalloc((void **)&A->i, s);
	if(A->i == NULL) {
		// TODO: deinit base class
		free(A);
		return NULL;
	}
	A->j = A->i + nnz;
	A->v = (ELT *)(A->j + n + 1);

	return (matrix_t *)A;
}

matrix_t *matrix_gpu_create_coo(cusparseHandle_t cusparse, linalg_elt_type_t type, int ndims, const int *dims, int nnz);


/////////////////////////////// Clone //////////////////////////

static matrix_t *csc_clone(const matrix_t *base)
{
	const gpu_sparse_t *src;
	gpu_sparse_t *dst;

	int m, n, nnz;

	assert(base->ndims == 2);
	m = base->dims[0];
	n = base->dims[1];

	src = cast_const(base);
	nnz = src->nnz;
	dst = cast(matrix_gpu_create_csc(src->handle, base->type, m, n, nnz));

	// Copy data
	cudaMemcpy(dst->i, src->i, (nnz + n + 1) * sizeof(int) + nnz * sizeof(ELT), cudaMemcpyDeviceToDevice);

	return (matrix_t *)dst;
}

static matrix_t *csr_clone(const matrix_t *base)
{
	const gpu_sparse_t *src;
	gpu_sparse_t *dst;

	int m, n, nnz;

	assert(base->ndims == 2);
	m = base->dims[0];
	n = base->dims[1];

	src = cast_const(base);
	nnz = src->nnz;
	dst = cast(matrix_gpu_create_csr(src->handle, base->type, m, n, nnz));

	// Copy data
	cudaMemcpy(dst->i, src->i, (nnz + m + 1) * sizeof(int) + nnz * sizeof(ELT), cudaMemcpyDeviceToDevice);

	return (matrix_t *)dst;
}

static matrix_t *coo_clone(const matrix_t *base);

/////////////////////////////// Destructors ////////////////////////

static void destroy(matrix_t *base)
{
	gpu_sparse_t *A;

	A = cast(base);

  cusparseDestroyMatDescr(A->desc);
	cudaFree(A->i);
	free(base);
}

/////////////////////////// Format Conversion ////////////////////////

static matrix_t *csc2full(const matrix_t *base);

static matrix_t *csc2csr(const matrix_t *base)
{
	const gpu_sparse_t *csc;
	gpu_sparse_t *csr;
	int m, n;

	m = base->dims[0];
	n = base->dims[1];

	assert(base->ndims == 2);
	csc = cast_const(base);
	csr = cast(matrix_gpu_create_csr(csc->handle, base->type, m, n, csc->nnz));

  CUSPARSE_FUNCTION(csr2csc)(csc->handle, m, n,
														 csc->v, csc->j, csc->i,
														 csr->v, csr->j, csr->i,
														 1, CUSPARSE_INDEX_BASE_ZERO);
}

static matrix_t *csc2coo(const matrix_t *base);

static matrix_t *csr2full(const matrix_t *base);

static matrix_t *csr2csc(const matrix_t *base)
{
	const gpu_sparse_t *csr;		
	gpu_sparse_t *csc;
	int m, n;

	m = base->dims[0];
	n = base->dims[1];

	assert(base->ndims == 2);
	csr = cast_const(base);
	csc = cast(matrix_gpu_create_csc(csr->handle, base->type, m, n, csr->nnz));

  CUSPARSE_FUNCTION(csr2csc)(csr->handle, m, n,
														 csr->v, csr->i, csr->j,
														 csc->v, csc->i, csc->j,
														 1, CUSPARSE_INDEX_BASE_ZERO);
}

static matrix_t *csr2coo(const matrix_t *base);
static matrix_t *coo2full(const matrix_t *base);

static matrix_t *coo2full(const matrix_t *base);
static matrix_t *coo2csr(const matrix_t *base);
static matrix_t *coo2csc(const matrix_t *base);

/////////////////////////////// NNZ ///////////////////////

static int nnz(const matrix_t *base)
{
	const gpu_sparse_t *A;

	A = cast_const(base);

	return A->nnz;
}

//////////////////////////////// vtables //////////////////////

static const matrix_vtable_t csr_vtable = {
	.clone = csr_clone,
	.destroy = destroy,

	.to_csr = csr_clone,
	.to_csc = csr2csc,

  .to_gpu = csr_clone,

  .nnz = nnz
};

static const matrix_vtable_t csc_vtable = {
	.clone = csc_clone,
	.destroy = destroy,

	.to_csr = csc2csr,
	.to_csc = csc_clone,

  .to_gpu = csc_clone,

	.nnz = nnz
};


