#include "../vector_internal.h"

#include <assert.h>
#include <stdlib.h>

#include <cuda_runtime.h>
#include <cublas.h>

#include "convert.h"
#include "type_defs.h"

static const vector_vtable_t vtable;

typedef struct
{
  vector_t base;
  ELT *data;
  
} gpu_vector_t;

static inline gpu_vector_t *cast(vector_t *base)
{
  gpu_vector_t *v;

  assert(base->v == &vtable);
  assert(matdesc_computation_unit_get(&base->desc) == LINALG_GPU);
  v = (gpu_vector_t *)base;

  return v;
}

static inline const gpu_vector_t *cast_const(const vector_t *base)
{
  const gpu_vector_t *v;

  assert(base->v == &vtable);
  assert(matdesc_computation_unit_get(&base->desc) == LINALG_GPU);
  v = (const gpu_vector_t *)base;

  return v;
}

//////////////////////// Constructors/Destructor ////////////////////

vector_t *vector_gpu_create(const matdesc_t *desc, int length)
{
  gpu_vector_t *v;

  v = (gpu_vector_t *)malloc(sizeof(gpu_vector_t));
  if(v == NULL) {
    return NULL;
  }

  // Initialize base class
  v->base.v = &vtable;
  v->base.length = length;
  matdesc_assign(&v->base.desc, desc);
  matdesc_computation_unit_set(&v->base.desc, LINALG_GPU);

  // Allocate data
  cudaMalloc((void **)&v->data, sizeof(ELT) * length);
  if(v->data == NULL) {
    matdesc_deinit(&v->base.desc);
    free(v);
    return NULL;
  }

  return (vector_t *)v;
}

static vector_t *clone(const vector_t *base)
{
  const gpu_vector_t *src;
  gpu_vector_t *dst;

  src = cast_const(base);
  dst = cast(vector_gpu_create(&base->desc, base->length));

	cudaMemcpy(dst->data, src->data, sizeof(ELT) * base->length, cudaMemcpyDeviceToDevice);

  return (vector_t *)dst;
}

static linalg_status_t destroy(vector_t *base)
{
  gpu_vector_t *v;

  v = cast(base);

  cudaFree(v->data);
  matdesc_deinit(&base->desc);
  free(base);

  return LINALG_OK;
}

static linalg_status_t resize(vector_t *base, int new_length)
{
  gpu_vector_t *v;
  ELT *new_data;

  v = cast(base);

  // Allocate new data array
  cudaMalloc((void **)&new_data, new_length * sizeof(ELT));
  if(new_data == NULL) {
    return LINALG_ALLOCATION_ERROR;
  }

  // Copy old data into new array
	cudaMemcpy(new_data, v->data, sizeof(ELT) * base->length, cudaMemcpyDeviceToDevice);

  // Switch to new array
  cudaFree(v->data);
  v->data = new_data;
  base->length = new_length;

  return LINALG_OK;
}

////////////////////////////// Assignment ////////////////////////

static linalg_status_t to_plain(const vector_t *base, plain_vector_t *plain)
{
  const gpu_vector_t *v;
  linalg_elt_type_t v_type, plain_type;
  int n;
  void *convert;
  linalg_status_t s;

  // Set up vectors
  assert(base->length == plain->base.length);
  n = base->length;

  v = cast_const(base);
  v_type = matdesc_get_elt_type(&base->desc);
  plain_type = matdesc_get_elt_type(&plain->base.desc);

  // Copy data
  // TODO: potential issues with different complex types
  if(v_type == plain_type) {
    cudaMemcpy(plain->data, v->data, gpu_type_length(n, v_type), cudaMemcpyDeviceToHost);
  }
  else {
    // Create intermediate vector for converting types
    cudaMalloc(&convert, gpu_type_length(n, plain_type));
    if(convert == NULL) {
      return LINALG_ALLOCATION_ERROR;
    }

    // Convert to plain type
    s = gpu_convert(n, plain_type, convert, v_type, v->data);
    if(s != LINALG_OK) {
      cudaFree(convert);
      return s;
    }

    // Copy into plain vector
    cudaMemcpy(plain->data, convert, gpu_type_length(n, plain_type), cudaMemcpyDeviceToHost);

    cudaFree(convert);
  }

  return LINALG_OK;
}

static linalg_status_t from_plain(vector_t *base, const plain_vector_t *plain)
{
  gpu_vector_t *v;
  linalg_elt_type_t v_type, plain_type;
  int n;
  void *convert;
  linalg_status_t s;

  // Set up vectors
  assert(base->length == plain->base.length);
  n = base->length;

  v = cast(base);
  v_type = matdesc_get_elt_type(&base->desc);
  plain_type = matdesc_get_elt_type(&plain->base.desc);

  // Copy data
  // TODO: potential issues with different complex types
  if(v_type == plain_type) {
    cudaMemcpy(v->data, plain->data, gpu_type_length(n, v_type), cudaMemcpyHostToDevice);
  }
  else {
    // Create intermediate vector for converting types
    cudaMalloc(&convert, gpu_type_length(n, plain_type));
    if(convert == NULL) {
      return LINALG_ALLOCATION_ERROR;
    }

    // Copy from plain vector
    cudaMemcpy(convert, plain->data, gpu_type_length(n, plain_type), cudaMemcpyHostToDevice);

    // Convert from plain type
    s = gpu_convert(n, v_type, v->data, plain_type, convert);
    if(s != LINALG_OK) {
      cudaFree(convert);
      return s;
    }

    cudaFree(convert);
  }

  return LINALG_OK;
}

static linalg_status_t assign(vector_t *dst_base, const vector_t *src_base)
{
  linalg_status_t s;
  linalg_elt_type_t dst_type, src_type;
  gpu_vector_t *dst;
  const gpu_vector_t *src;

  // Set up vectors
  assert(dst_base->length == src_base->length);

  dst = cast(dst_base);
  src = cast_const(src_base);

  dst_type = matdesc_get_elt_type(&dst_base->desc);
  src_type = matdesc_get_elt_type(&src_base->desc);

  // Copy data
  return gpu_convert(dst_base->length, dst_type, dst->data, src_type, src->data);
}

//////////////////////////////// Norms ///////////////////////////

static double norm1(const vector_t *base);

static float norm1f(const vector_t *base)
{
  int n;
  const gpu_vector_t *v;

  n = base->length;
  v = cast_const(base);

  #ifdef DOUBLE
    return (float)(norm1(base));
  #else
    #ifdef COMPLEX
      return cublasScasum(n, v->data, 1);
    #else
      return cublasSasum(n, v->data, 1);
    #endif 
  #endif
}

static double norm1(const vector_t *base)
{
  int n;
  const gpu_vector_t *v;

  n = base->length;
  v = cast_const(base);

  #ifdef DOUBLE
    #ifdef COMPLEX
      return cublasDzasum(n, v->data, 1);
    #else
      return cublasDasum(n, v->data, 1);
    #endif 
  #else
    return (double)(norm1f(base));
  #endif
}

static double norm2(const vector_t *base);

static float norm2f(const vector_t *base)
{
  int n;
  const gpu_vector_t *v;

  n = base->length;
  v = cast_const(base);

  #ifdef DOUBLE
    return (float)(norm2(base));
  #else
    #ifdef COMPLEX
      return cublasScnrm2(n, v->data, 1);
    #else
      return cublasSnrm2(n, v->data, 1);
    #endif 
  #endif
}

static double norm2(const vector_t *base)
{
  int n;
  const gpu_vector_t *v;

  n = base->length;
  v = cast_const(base);

  #ifdef DOUBLE
    #ifdef COMPLEX
      return cublasDznrm2(n, v->data, 1);
    #else
      return cublasDnrm2(n, v->data, 1);
    #endif 
  #else
    return (double)(norm2f(base));
  #endif
}

///////////////////////////// Minimum/Maximum ////////////////////////

static int iamax(const vector_t *base)
{
  int n;
  const gpu_vector_t *v;

  n = base->length;
  v = cast_const(base);

  #ifdef DOUBLE
    #ifdef COMPLEX
      return cublasIzamax(n, v->data, 1)-1;
    #else
      return cublasIdamax(n, v->data, 1)-1;
    #endif 
  #else
    #ifdef COMPLEX
      return cublasIcamax(n, v->data, 1)-1;
    #else
      return cublasIsamax(n, v->data, 1)-1;
    #endif 
  #endif
}


static int iamin(const vector_t *base)
{
  int n;
  const gpu_vector_t *v;

  n = base->length;
  v = cast_const(base);

  #ifdef DOUBLE
    #ifdef COMPLEX
      return cublasIzamin(n, v->data, 1)-1;
    #else
      return cublasIdamin(n, v->data, 1)-1;
    #endif 
  #else
    #ifdef COMPLEX
      return cublasIcamin(n, v->data, 1)-1;
    #else
      return cublasIsamin(n, v->data, 1)-1;
    #endif 
  #endif
}

///////////////////////////// Element Access ////////////////////


static doublecomplex elt(const vector_t *base, int i)
{
  const gpu_vector_t *v;

  doublecomplex z;
  complex c;

  v = cast_const(base);

  #ifdef DOUBLE
    #ifdef COMPLEX
      cudaMemcpy(&z, v->data + i, sizeof(doublecomplex), cudaMemcpyDeviceToHost);
    #else
      cudaMemcpy(&z.r, v->data + i, sizeof(double), cudaMemcpyDeviceToHost);
      z.i = 0;
    #endif
  #else
    #ifdef COMPLEX
      cudaMemcpy(&c, v->data + i, sizeof(complex), cudaMemcpyDeviceToHost);
    #else
      cudaMemcpy(&c.r, v->data + i, sizeof(float), cudaMemcpyDeviceToHost);
    #endif 

    z.r = (double)c.r;
    #ifdef COMPLEX
      z.i = (double)c.i;
    #else
      z.i = 0;
    #endif
  #endif

  return z;
}

/////////////////////////////// vtable //////////////////////////

static const vector_vtable_t vtable = {
  .clone = clone,
  .destroy = destroy,
  .resize = resize,

  .to_plain = to_plain,
  .from_plain = from_plain,
  .assign = assign,

  .norm1f = norm1f,
  .norm1 = norm1,
  .norm2f = norm2f,
  .norm2 = norm2,

  .iamax = iamax,
  .iamin = iamin,

  .elt = elt
};


