#include "plain.h"
#include "../vector_internal.h"

#include "type_defs.h"

#include <stdlib.h>
#include <memory.h>
#include <assert.h>

static const vector_vtable_t vtable;

static inline plain_vector_t *cast(vector_t *base)
{
  plain_vector_t *v;

  assert(base->v == &vtable);
  assert(matdesc_computation_unit_get(&base->desc) == LINALG_GPU);
  v = (plain_vector_t *)base;

  return v;
}

static inline const plain_vector_t *cast_const(const vector_t *base)
{
  const plain_vector_t *v;

  assert(base->v == &vtable);
  assert(matdesc_computation_unit_get(&base->desc) == LINALG_GPU);
  v = (const plain_vector_t *)base;

  return v;
}

vector_t *vector_plain_create(const matdesc_t *desc, int length)
{
  plain_vector_t *v;

  v = (plain_vector_t *)malloc(sizeof(plain_vector_t));
  if(v == NULL) {
    return NULL;
  }

  // Initialize base class
  v->base.v = &vtable;
  v->base.length = length;
  matdesc_assign(&v->base.desc, desc);
  matdesc_computation_unit_set(&v->base.desc, LINALG_PLAIN);

  // Allocate data
  v->data = malloc(sizeof(ELT) * length);
  if(v->data == NULL) {
    matdesc_deinit(&v->base.desc);
    free(v);
    return NULL;
  }

  return (vector_t *)v;
}

static vector_t *clone(const vector_t *base)
{
  const plain_vector_t *src;
  plain_vector_t *dst;

  src = cast_const(base);
  dst = cast(vector_plain_create(&base->desc, base->length));

	memcpy(dst->data, src->data, sizeof(ELT) * base->length);

  return (vector_t *)dst;
}

static linalg_status_t destroy(vector_t *base)
{
  plain_vector_t *v;

  v = cast(base);

  free(v->data);
  matdesc_deinit(&base->desc);
  free(base);

  return LINALG_OK;
}

static linalg_status_t resize(vector_t *base, int new_length)
{
  plain_vector_t *v;
  ELT *new_data;

  v = cast(base);

  v->data = realloc(v->data, new_length * sizeof(ELT));
  base->length = new_length;

  return LINALG_OK;
}

////////////////////////////// Assignment ////////////////////////

static linalg_status_t assign(vector_t *dst_base, const vector_t *src_base)
{
  linalg_status_t s;
  linalg_elt_type_t dst_type, src_type;
  plain_vector_t *dst;
  const plain_vector_t *src;

  // Set up vectors
  assert(dst_base->length == src_base->length);

  dst = cast(dst_base);
  src = cast_const(src_base);

  dst_type = matdesc_get_elt_type(&dst_base->desc);
  src_type = matdesc_get_elt_type(&src_base->desc);

  // Copy data
  return plain_convert(dst_base->length, dst_type, dst->data, src_type, src->data);
}

static linalg_status_t to_plain(const vector_t *base, plain_vector_t *plain)
{
  return assign((vector_t *)plain, base);
}

static linalg_status_t from_plain(vector_t *base, const plain_vector_t *plain)
{
  return assign(base, (const vector_t *)plain);
}
