#include "vector_internal.h"

#include <stdlib.h>
#include <assert.h>
#include <math.h>

#include "scalar.h"

vector_t *vector_clone(const vector_t *v)
{
  if(v->v->clone == NULL) {
    return NULL;
  }
  return v->v->clone(v);
}

linalg_status_t vector_destroy(vector_t *v)
{
  if(v == NULL) {
    return;
  }
  return v->v->destroy(v);
}

linalg_status_t vector_assign(vector_t *dst, const vector_t *src)
{
  linalg_computation_unit_t dst_cu, src_cu;
  linalg_status_t s;

  // Resize so vectors are the same size
  if(dst->length != src->length) {
    s = vector_resize(dst, src->length);
    if(s != LINALG_OK) {
      return s;
    }
  }

  src_cu = matdesc_computation_unit_get(&src->desc);
  dst_cu = matdesc_computation_unit_get(&dst->desc);

  if(src_cu == dst_cu && dst->v->assign != NULL) {
    return dst->v->assign(dst, src);
  }
  else if(src_cu == LINALG_PLAIN) {
    return dst->v->from_plain(dst, (const plain_vector_t *)src);
  }
  else if(dst_cu == LINALG_PLAIN) {
    return src->v->to_plain(src, (plain_vector_t *)dst);
  }
  else {
    // Convert between computation units through an intermediatary
    // plain vector
    // TODO
    assert(0);
  }
}

int vector_length(const vector_t *v)
{
  return v->length;
}

linalg_status_t vector_resize(vector_t *v, int new_length)
{
  linalg_status_t s;

  if(v->v->resize == NULL) {
    return LINALG_OPERATION_NOT_SUPPORTED;
  }
  
  s = v->v->resize(v, new_length);
  assert(s != LINALG_OK || v->length == new_length);

  return s;
}

const matdesc_t *vector_description(const vector_t *v)
{
  return &v->desc;
}

/////////////////////////// Norms ////////////////////////

float vector_norm1f(const vector_t *v)
{
  assert(v->v->norm1f != NULL);
  return v->v->norm1f(v);
}

double vector_norm1(const vector_t *v)
{
  assert(v->v->norm1 != NULL);
  return v->v->norm1(v);
}


float vector_norm2f(const vector_t *v)
{
  assert(v->v->norm2f != NULL);
  return v->v->norm2f(v);
}

double vector_norm2(const vector_t *v)
{
  assert(v->v->norm2 != NULL);
  return v->v->norm2(v);
}

float vector_norminftyf(const vector_t *v)
{
  int i;
  complex c;

  i = vector_iamax(v);
  c = vector_celt(v, i);

  return fabsf(c.r) + fabsf(c.i);
}

double vector_norminfty(const vector_t *v)
{
  int i;
  doublecomplex z;

  i = vector_iamax(v);
  z = vector_zelt(v, i);

  return fabs(z.r) + fabs(z.i);
}

///////////////////////// Minimum/Maximum //////////////////////

int vector_iamin(const vector_t *v)
{
  assert(v->v->iamin != NULL);
  return v->v->iamin(v);
}

int vector_iamax(const vector_t *v)
{
  assert(v->v->iamax != NULL);
  return v->v->iamax(v);
}

////////////////////////////// Element Access /////////////////////////

complex vector_celt(const vector_t *v, int i)
{
  return ZtoC(vector_zelt(v, i));
}

doublecomplex vector_zelt(const vector_t *v, int i)
{
  assert(v->v->elt != NULL);
  assert(i >= 0 && i < v->length);
  return v->v->elt(v, i);
}

///////////////////////// Vector Math ////////////////////////

/////////////// Dot /////////////////

complex vector_cdot(const vector_t *u, const vector_t *v)
{
  return ZtoC(vector_zdot(u, v));
}

doublecomplex vector_zdot(const vector_t *u, const vector_t *v)
{
  // TODO: check math compatibility
}

/////////////// Scale /////////////////

void vector_sscale(float s, vector_t *v)
{
  vector_zscale(StoZ(s), v);
}

void vector_dscale(double s, vector_t *v)
{
  vector_zscale(DtoZ(s), v);
}

void vector_cscale(complex s, vector_t *v)
{
  vector_zscale(CtoZ(s), v);
}

void vector_zscale(doublecomplex s, vector_t *v)
{
  assert(v->v->scale != NULL);
  v->v->scale(s, v);
}

/////////////// a * x + y /////////////////

void vector_saxpy(float alpha, const vector_t *x, vector_t *y)
{
  vector_zaxpy(StoZ(alpha), x, y);
}

void vector_daxpy(double alpha, const vector_t *x, vector_t *y)
{
  vector_zaxpy(DtoZ(alpha), x, y);
}

void vector_caxpy(complex alpha, const vector_t *x, vector_t *y)
{
  vector_zaxpy(CtoZ(alpha), x, y);
}

void vector_zaxpy(doublecomplex alpha, const vector_t *x, vector_t *y)
{
  // TODO: check math compatibiliy

  assert(y->v->axpy != NULL);
  y->v->axpy(alpha, x, y);
}
