#include "project.h"

#include <stdlib.h>
#include <stdio.h>
#include <time.h>

#include <cuda_runtime.h>
#include <cublas.h>

#include "runparams.h"

#ifdef __cplusplus
extern "C" {
#endif

void boxproject(int n, double *x, const double *l, const double *u)
{
  int i;

  for(i = 0; i < n; i++) {
    if(x[i] < l[i]) {
      x[i] = l[i];
    }
    else if(x[i] > u[i]) {
      x[i] = u[i];
    }
  }
}

#ifdef __cplusplus
}
#endif

// Device code
__global__ void boxproject_kernel(
  int n, double *x, const double *l, const double *u)
{
  int i;

  // Get index
  i = blockDim.x * blockIdx.x + threadIdx.x;
  if(i >= n) {
    return;
  }

  // Project
  if(x[i] < l[i]) 
{
    x[i] = l[i];
  }
  else if(x[i] > u[i]) {
    x[i] = u[i];
  }

 

}


#ifdef __cplusplus
extern "C" {
#endif

void boxproject_gpu(int n, double *x, const double *l, const double *u, const struct cudaDeviceProp *devprops)
{
	runparams_t p;

	p = runparams(n, devprops);

  while(n > 0) {
    boxproject_kernel<<<p.blocks, p.threads>>>(n, x, l, u);

    x += p.stride;
    l += p.stride;
    u += p.stride;

    n -= p.stride;
  }
}


#ifdef __cplusplus
}
#endif

#ifdef __cplusplus
extern "C" {
#endif

void boxproject_direction(int n, double *d, const double *x, const double *l, const double *u)
{
  int i;

  for(i = 0; i < n; i++) {
    if(x[i] <= l[i] && d[i] < 0) {
      d[i] = 0;
    }
    else if(x[i] >= u[i] && d[i] > 0) {
      d[i] = 0;
    }
  }
}

#ifdef __cplusplus
}
#endif


__global__ void project_direction_kernel(
  int n, double *d, const double *x, const double *l, const double *u)
{
  int i;

  // Get index
  i = blockDim.x * blockIdx.x + threadIdx.x;
  if(i >= n) {
    return;
  }

  // Project
  if(x[i] <= l[i] && d[i] < 0) {
    d[i] = 0;
  }
  else if(x[i] >= u[i] && d[i] > 0) {
    d[i] = 0;
  }
}

#ifdef __cplusplus
extern "C" {
#endif

void boxproject_direction_gpu(int n, double *d, const double *x, const double *l, const double *u, struct cudaDeviceProp *devprops)
{
	runparams_t p;

	p = runparams(n, devprops);

  while(n > 0) {
    project_direction_kernel<<<p.blocks, p.threads>>>(n, d, x, l, u);

    d += p.stride;

    x += p.stride;
    l += p.stride;
    u += p.stride;

    n -= p.stride;
  }
}

#ifdef __cplusplus
}
#endif

#ifdef __cplusplus
extern "C" {
#endif

double boxproject_breakpoint(int n, const double *x, const double *d, const double *l, const double *u)
{
  int i;
  double b, bi;

  b = INFINITY;

  for(i = 0; i < n; i++) {
    if(d[i] > 0) {
      bi = (u[i] - x[i]) / d[i];
      if(bi < b) {
        b = bi;
      }
    }
    else if(d[i] < 0) {
      bi = (l[i] - x[i]) / d[i];
      if(bi < b) {
        b = bi;
      }
    }
  }

  return b;
}

#ifdef __cplusplus
}
#endif

__global__ void breakpoint_dims_kernel(
  int n, double *b, const double *x, const double *d, const double *l, const double *u)
{
  int i;

  // Get index
  i = blockDim.x * blockIdx.x + threadIdx.x;
  if(i >= n) {
    return;
  }

  if(d[i] < 0) {
    b[i] = (u[i] - x[i]) / d[i];
  }
  else if(d[i] > 0) {
    b[i] = (l[i] - x[i]) / d[i];
 }
  else {
    b[i] = (1.0/0.0);
  }
}

extern "C" double boxproject_breakpoint_gpu(int n, const double *x, const double *d, const double *l, const double *u, double *b, struct cudaDeviceProp *devprops)
{
	runparams_t p;

	p = runparams(n, devprops);

  int mini;
  double minb;

  // Get the breakpoints for each dimension
  while(n > 0) {
    breakpoint_dims_kernel<<<p.blocks, p.threads>>>(n, b, x, d, l, u);

    x += p.stride;
    d += p.stride;
    l += p.stride;
    u += p.stride;
    b += p.stride;

    n -= p.stride;
  }

  // Find the least breakpoint
  mini = cublasIdamin(n, x, 1);
  cudaMemcpy(&minb, b + mini, sizeof(double), cudaMemcpyDeviceToHost);

  return minb;
}
