#include "set.h"

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <assert.h>

#include <cuda_runtime.h>

static __global__ void symm_diff_kernel(int n, bool *diff, const bool *S, const bool *T)
{
  int i;

  // Get index
  i = blockDim.x * blockIdx.x + threadIdx.x;
  if(i >= n) {
    return;
  }

  diff[i] = (S[i] != T[i]);
}


void set_symm_diff(
  int n, bool *diff, const bool *S, const bool *T,
  struct cudaDeviceProp *devprops)
{
  int threads, blocks;
  int stride;

  threads = devprops->maxThreadsPerBlock;
  blocks = devprops->maxGridSize[0];
  stride = threads * blocks;

  while(n > 0) {
    symm_diff_kernel<<<blocks, threads>>>(n, diff, S, T);

    diff += stride;

    S += stride;
    T += stride;

    n -= stride;
  }
}


extern "C" void print_set(int n, const bool *x, int group, const char *spacing);
extern "C" void print_set_gpu(int n, const bool *x, int group, const char *spacing);

void print_ints(int n, const int *x, int group, const char *spacing)
{
  int i;

  if(spacing == NULL) {
    spacing = "";
  }

  for(i = 0; i < n; i++) {
    if(group > 0 && i != 0 && i % group == 0) {
      fputc(' ', stdout);
    }

    printf("%2d", x[i]);

    fputs(spacing, stdout);
  }
  fputc('\n', stdout);
}


void print_ints_gpu(int n, const int *x, int group, const char *spacing)
{
  int *cpu;

  cpu = (int *)malloc(n * sizeof(int));
  cudaMemcpy(cpu, x, n * sizeof(int), cudaMemcpyDeviceToHost);
  print_ints(n, cpu, group, spacing);
  free(cpu);
}

static __global__ void or_accum_kernel(int n, bool *dst, bool *src, bool *block_accum)
{
  int i, base;
  bool *swap;

  //int uid = blockDim.x * blockIdx.x + threadIdx.x;

  // Get range handled by this block
  base = blockDim.x * blockIdx.x;
  if(2*base > n) {
    block_accum[blockIdx.x] = false;
    return;
  }

  dst += base;
  src += 2 * base;
  n -= 2 * base;

  if(2*blockDim.x < n) {
    n = 2*blockDim.x;
  }

  // Get myself...
  i = threadIdx.x;

  // thread i handles 2*i, 2*i+1

  do {
    __syncthreads();

    // Calculate a pairwise or
    if(2*i >= n) {
      return;
    }
    else if(2*i == n-1) {
      dst[i] = src[2*i];
    }
    else {
      dst[i] = src[2*i] || src[2*i+1];
    }

    // Set up next problem
    swap = dst;
    dst = src;
    src = swap;

    n = (n + 1)/2;
  } while(n > 1);

  block_accum[blockIdx.x] = src[i];
}

bool set_or(int n, const bool *S, bool *work, bool *host_work, struct cudaDeviceProp *devprops)
{
  int threads, blocks, stride, i;
  bool *work1, *work2, *block_accum;

  // Get the number of necessary kernel calls
  threads = devprops->maxThreadsPerBlock;
  //threads = 4;
  if(n < threads) {
    threads = n;
    blocks = 1;
  }
  blocks = devprops->maxGridSize[0];
  stride = threads * blocks;
  if(stride > (n+1)/2) {
    blocks = ((n+1)/2 + threads - 1) /  threads;
    stride = threads * blocks;
  }

  // Determine locations of workspaces
  work1 = work;
  work2 = work + n;
  block_accum = work + n + (n+1)/2;
  assert(work + 2*n - block_accum > blocks);

  // Copy input to work array
  cudaMemcpy(work1, S, n * sizeof(bool), cudaMemcpyDeviceToDevice);

  //printf("blocks = %d, threads = %d, stride = %d\n", blocks, threads, stride);

  while(n > 0) {
    or_accum_kernel<<<blocks, threads>>>(n, work2, work1, block_accum);

    // Check if any blocks returned true
    cudaMemcpy(host_work, block_accum, blocks * sizeof(bool), cudaMemcpyDeviceToHost);
    for(i = 0; i < blocks; i++) {
      if(host_work[i]) {
        return true;
      }
    }

    S += 2*stride;
    n -= 2*stride;

    break;
  }

  return false;
}

bool sets_equal(
  int n, const bool *S, const bool *T,
  bool *work, bool *host_work, struct cudaDeviceProp *devprops)
{
  set_symm_diff(n, work, S, T, devprops);

  return !set_or(n, work, work + n, host_work, devprops); 
}
