#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <time.h>
#include <memory.h>

#include <cusparse.h>
#include <cublas.h>
#include <cuda_runtime.h>

#include <boxopt/boxqp.h>

static gpu_matf_t test_Q(int n)
{
  gpu_matf_t Q;
  int nnz;
  int i;

  Q.format = GPU_MAT_CSR;
  
  // Allocate
  Q.m = Q.n = n;
  Q.i = malloc(sizeof(int) * (n + 1));

  nnz = n;
  Q.j = malloc(sizeof(int) * nnz);
  Q.v = malloc(sizeof(float) * nnz);

  // Fill
  nnz = 0;
  for(i = 0; i < n+1; i++) {
    Q.i[i] = nnz++;
  }

  for(i = 0; i < n; i++) {
    Q.j[i] = i;
    Q.v[i] = 1;
  }

  return Q;
}

void frandf(int n, float *x)
{
  int i;

  for(i = 0; i < n; i++) {
    x[i] = rand() / (float)RAND_MAX;
  }
}

static gpu_boxqpf_t test_BoxQP(int n, float *x)
{
  gpu_boxqpf_t boxqp;
  int i;

  // Allocate
  boxqp.c = (float *)malloc(3 * n * sizeof(float));
  boxqp.l = boxqp.c + n;
  boxqp.u = boxqp.l + n;

  // Set
  frandf(3 * n, boxqp.c);
  for(i = 0; i < n; i++) {
    boxqp.u[i] += boxqp.l[i];
    boxqp.l[i] *= -1;
    boxqp.c[i] = 2 * boxqp.c[i] - 1;
  }

  boxqp.Q = test_Q(n);

  memcpy(x, boxqp.l, sizeof(float) * n);

  return boxqp;
}

static gpu_boxqpf_t copy_BoxQP(const gpu_boxqpf_t *cpu_boxqp, float **outx, const float *cpu_x)
{
  gpu_boxqpf_t boxqp;
  size_t svec;
  int n, nnz;
  float *x;

  n = cpu_boxqp->Q.n;
  svec = n * sizeof(float);

  // Copy vectors
  cudaMalloc((void **)outx, 4 * svec);
  x = *outx;
  boxqp.c = x + n;
  boxqp.l = boxqp.c + n;
  boxqp.u = boxqp.l + n;

  // TODO: memcpy
  
  // Copy Q
  boxqp.Q.format = cpu_boxqp->Q.format;
  boxqp.Q.m = cpu_boxqp->Q.m;
  boxqp.Q.n = cpu_boxqp->Q.n;

  assert(cpu_boxqp->Q.format == GPU_MAT_CSR);
  nnz = cpu_boxqp->Q.i[n] - cpu_boxqp->Q.i[0];
  //cudaMalloc((void **)boxqp.Q.i, boxqp.Q.

  // TODO: allocate and memcpy

  return boxqp;
}

static void free_gpu_BoxQP(gpu_boxqpf_t *boxqp, float *x)
{
  cudaFree(x);
}

static void cublasShutdownNoReturn()
{
  cublasShutdown();
}

static cusparseHandle_t cusparse_handle;

static void cusparseShutdownNoReturn()
{
  cusparseDestroy(cusparse_handle);
}

int main(int argc, char **argv)
{
  int i, n;
  gpu_boxqpf_t boxqp, gpu_boxqp;
  gpfr_options_t opts;
  float *x, *gpu_x;

  srand(time(NULL));

  // Set up CUBLAS
  if(cublasInit() != CUBLAS_STATUS_SUCCESS) {
    exit(1);
  }
  atexit(&cublasShutdownNoReturn);

  // Set up CUSPARSE
  if(cusparseCreate(&cusparse_handle) != CUSPARSE_STATUS_SUCCESS) {
    exit(1);
  }
  atexit(&cusparseShutdownNoReturn);

  // Construct test BoxQP
  n = 16*1024;
  x = malloc(sizeof(float) * n);
  boxqp = test_BoxQP(n, x);
  gpu_boxqp = copy_BoxQP(&boxqp, &gpu_x, x);

  // Solve
  opts = gpfr_default_options;
  opts.max_iters = 10;
  gpu_boxqp_grad_projf(gpu_x, &gpu_boxqp, &opts, cusparse_handle);

  //int err;
  //err = cudaGetLastError();
  //printf("%s:%d - %s\n", __FILE__, __LINE__, cudaGetErrorString(err));

  // Print solution
  //for(i = 0; i < (n < 16 ? n : 16); i++) {
  //  printf("%e %e [%e, %e]\n", x[i], boxqp.c[i], boxqp.l[i], boxqp.u[i]);
  //}

  // Cleanup
  free_gpu_BoxQP(&gpu_boxqp, gpu_x);
  // TODO
}

/*
int main(int argc, char **argv)
{
  int i, n;
  float *x, *l, *u, *p, *p2;
  float *x_gpu, *l_gpu, *u_gpu;
  size_t svec;

  struct timeval timer;
  double gpu_secs, cpu_secs;

  cudaError_t err;
  int device;
  struct cudaDeviceProp devprops;

  cudaGetDevice(&device);
  cudaGetDeviceProperties(&devprops, device);

  printf("threads per block: %d\n", devprops.maxThreadsPerBlock);
  printf("max grid dim: %d %d %d\n", devprops.maxGridSize[0], devprops.maxGridSize[1], devprops.maxGridSize[2]);
  printf("max memory: %lu\n", devprops.totalGlobalMem);

  srand(time(NULL));

  n = 16 * 1024 * 1024;

  // Allocation
  svec = sizeof(float) * n;
  x = (float *)malloc(5 * svec);
  l = x + n;
  u = l + n;
  p = u + n;
  p2 = p + n;

  err = cudaGetLastError();
  printf("%s\n", cudaGetErrorString(err));

  cudaMalloc(&x_gpu, 3 * svec);
  l_gpu = x_gpu + n;
  u_gpu = l_gpu + n;

  err = cudaGetLastError();
  printf("%s:%d - %s\n", __FILE__, __LINE__, cudaGetErrorString(err));

#define CLEANUP \
  cudaFree(x_gpu); \
  free(x); \
  cudaThreadExit();

  if(x_gpu == NULL) {
    fprintf(stderr, "Allocation error.\n");
    CLEANUP;
    return 1;
  }

  // Initialize vectors
  frandf(3*n, x);
  for(i = 0; i < n; i++) {
    u[i] += l[i];
  }

  // Copy to GPU
  cudaMemcpy(x_gpu, x, svec, cudaMemcpyHostToDevice);
  cudaMemcpy(l_gpu, l, svec, cudaMemcpyHostToDevice);
  cudaMemcpy(u_gpu, u, svec, cudaMemcpyHostToDevice);

  // Perform projection on GPU
  timer_start(&timer);
  boxprojectf_gpu(n, x_gpu, l_gpu, u_gpu, &devprops);

  err = cudaGetLastError();
  printf("%s:%d - %s\n", __FILE__, __LINE__, cudaGetErrorString(err));

  cudaThreadSynchronize();
  gpu_secs = timer_interval(&timer);

  err = cudaGetLastError();
  printf("%s:%d - %s\n", __FILE__, __LINE__, cudaGetErrorString(err));

  // Copy result back to host
  cudaMemcpy(p, x_gpu, svec, cudaMemcpyDeviceToHost);

  err = cudaGetLastError();
  printf("%s:%d - %s\n", __FILE__, __LINE__, cudaGetErrorString(err));

  // Print
  if(n <= 32) {
    printf("On GPU:\n");
    for(i = 0; i < n; i++) {
      printf("%.3e %.3e %.3e = %.3e\n", l[i], x[i], u[i], p[i]);
    }
  }
  printf("GPU time: ");  time_print(stdout, gpu_secs); fputc('\n', stdout);

  // Repeat on CPU
  memcpy(p2, x, svec);
  timer_start(&timer);
  boxprojectf(n, p2, l, u);
  cpu_secs = timer_interval(&timer);

  if(n <= 32) {
    printf("\nOn CPU:\n");
    for(i = 0; i < n; i++) {
      printf("%.3e %.3e %.3e = %.3e\n", l[i], x[i], u[i], p2[i]);
    }
  }
  printf("CPU time: ");  time_print(stdout, cpu_secs); fputc('\n', stdout);

  for(i = 0; i < n; i++) {
    if(p[i] != p2[i]) {
      printf("Disagreement! %d: %g, %g\n", i, p[i], p2[i]);
      break;
    }
  }

  // Exit
  CLEANUP;
  return 0;
}           
*/
