#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <time.h>
#include <sys/time.h>
#include <assert.h>

#include <cuda_runtime.h>
#include <cusparse.h>

#include "../src/debug.h"

const int Ai[] = {0, 4, 7, 10, 14};

const int Ai_coo[] = {0, 0, 0, 0, 1, 1, 1, 2, 2, 2, 3, 3, 3, 3};
const int Aj[]    = {0, 1, 2, 3, 0, 1, 3, 0, 2, 3, 0, 1, 2, 3};
const double Av[] = {3.478624e+00, 4.429683e-01, 1.013380e+00, 8.248479e-01, 4.429683e-01, 2.415541e+00, 3.302211e-01, 1.013380e+00, 1.801608e+00, 7.709315e-02, 8.248479e-01, 3.302211e-01, 7.709315e-02, 1.232162e+00};

const double x[] = {-2.782585e-01, 7.706978e-01, 7.539797e-01, -1.217095e+00};

int main(int argc, char **argv)
{
  cusparseHandle_t cusparse_handle = NULL;
  int m, n, nnz;
  size_t Asize;

  int *Bi, *Bj, *Bi_coo;
  double *Bv;

  double *gpu_x, *Bx, *Ax;

  cusparseMatDescr_t desc;
  cusparseStatus_t s;

  cudaError_t err;

  // Sizes
  m = n = sizeof(Ai) / sizeof(int) - 1;
  nnz = sizeof(Aj) / sizeof(int);
  printf("m = %d, n = %d, nnz = %d\n", m, n, nnz);
  assert(nnz == Ai[m] - Ai[0]);

  if(cusparseCreate(&cusparse_handle) != CUSPARSE_STATUS_SUCCESS) {
    fprintf(stderr, "CUSPARSE startup error.");
    return 1;
  }

  // Allocate
  cudaAssertSuccess(cudaMalloc((void **)&Bi, (m + 1) * sizeof(int)));
  cudaAssertSuccess(cudaMalloc((void **)&Bj, nnz * sizeof(int)));
  cudaAssertSuccess(cudaMalloc((void **)&Bv, nnz * sizeof(double)));
  cudaAssertSuccess(cudaMalloc((void **)&Bi_coo, nnz * sizeof(double)));

  cudaAssertSuccess(cudaMemcpy(Bi_coo, Ai_coo, nnz * sizeof(int), cudaMemcpyHostToDevice));

  // Copy to GPU
  cudaAssertSuccess(cudaMemcpy(Bi, Ai, (m + 1) * sizeof(int), cudaMemcpyHostToDevice));
  cudaAssertSuccess(cudaMemcpy(Bj, Aj, nnz * sizeof(int), cudaMemcpyHostToDevice));
  cudaAssertSuccess(cudaMemcpy(Bv, Av, nnz * sizeof(double), cudaMemcpyHostToDevice));

  cudaAssertSuccess(cudaMalloc((void **)&gpu_x, (m + n) * sizeof(double)));
  Bx = gpu_x + m;

  cudaAssertSuccess(cudaMemcpy(gpu_x, x, m * sizeof(double), cudaMemcpyHostToDevice));

  Ax = (double *)malloc(n * sizeof(double));
  assert(Ax != NULL);

  cusparseAssertSuccess(cusparseCreateMatDescr(&desc));
  cusparseAssertSuccess(cusparseSetMatType(desc, CUSPARSE_MATRIX_TYPE_GENERAL));
  cusparseAssertSuccess(cusparseSetMatIndexBase(desc, CUSPARSE_INDEX_BASE_ZERO));

  print_ints_gpu(m+1, Bi);
  print_ints_gpu(nnz, Bj);
  print_vector_gpu(nnz, Bv);

  // Convert COO to CSR
  print_ints_gpu(nnz, Bi_coo);
  cusparseAssertSuccess(cusparseXcoo2csr(cusparse_handle, Bi_coo, nnz, m, Bi, CUSPARSE_INDEX_BASE_ZERO));

  cudaThreadSynchronize();
  err = cudaGetLastError();
  printf("err = %s\n", cudaGetErrorString(err));
  print_ints_gpu(m+1, Bi);

  // Perform multiplication
  print_vector_gpu(n, gpu_x);
  cudaThreadSynchronize();
  s = cusparseDcsrmv(cusparse_handle, CUSPARSE_OPERATION_NON_TRANSPOSE,
                     m, n,
                     1, desc, Bv, Bi, Bj, gpu_x,
                     1, Bx);
  cusparseAssertSuccess(s);
  cudaThreadSynchronize();
  printf("Qmv status = %d\n", (int)s);

  err = cudaGetLastError();
  printf("err = %s\n", cudaGetErrorString(err));

  // Print result
  print_vector_gpu(n, Bx);

  // Cleanup
  cusparseDestroyMatDescr(desc);
  free(Ax);
  cudaFree(Bx);
  cudaFree(gpu_x);
  cudaFree(Bi);
  cusparseDestroy(cusparse_handle);

  return 0;
}
